# -*- coding: utf-8 -*-

from django.db import models

from bitfield_manager import utils


class ParentBitfieldModelMixin(object):
    def set_flag(self, status_field, flag, save=False):
        status_value = getattr(self, status_field)
        status_value = utils.set_flag_field_for_status(status_value, flag)
        setattr(self, status_field, status_value)
        if save:
            self.save()
        return self

    def unset_flag(self, status_field, flag, save=False):
        status_value = getattr(self, status_field)
        status_value = utils.unset_flag_field_for_status(status_value, flag)
        setattr(self, status_field, status_value)
        if save:
            self.save()
        return self

    def force_status_refresh(self, related_models=None, search_depth=1):
        if not related_models:
            related_models = utils.get_all_related_bitfield_models(self.__class__, search_depth=search_depth)
        for i in related_models:
            parent_models = i.BitfieldMeta.parent_models
            for source, flag in parent_models:
                field = utils.get_field_from_source(source)
                qs = utils.get_django_query_string_from_source(source)
                related_count = i.objects.filter(**{qs: self.id}).count()
                status_value = getattr(self, field)
                if related_count == 0 and utils.is_flag_field_set_for_status(status_value, flag):
                    status_value = utils.unset_flag_field_for_status(status_value, flag)
                    setattr(self, field, status_value)
                    self.save(update_fields=[field])
                elif related_count > 0 and not utils.is_flag_field_set_for_status(status_value, flag):
                    status_value = utils.set_flag_field_for_status(status_value, flag)
                    setattr(self, field, status_value)
                    self.save(update_fields=[field])
        return self


class ChildBitfieldModelMixin(object):
    def save(self, *args, **kwargs):
        if not getattr(self, 'BitfieldMeta', None):
            super(ChildBitfieldModelMixin, self).save(*args, **kwargs)
            return
        parent_models = self.BitfieldMeta.parent_models

        for source, flag in parent_models:
            field = utils.get_field_from_source(source)
            parent_model = utils.get_parent_model(self, source)
            if not parent_model:
                # if not parent model skip
                continue
            else:
                utils.check_and_set_flag(parent_model, field, flag)
        super(ChildBitfieldModelMixin, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        if not getattr(self, 'BitfieldMeta', None):
            return super(ChildBitfieldModelMixin, self).delete(*args, **kwargs)

        parent_models = self.BitfieldMeta.parent_models
        for source, flag in parent_models:
            field = utils.get_field_from_source(source)
            parent_model = utils.get_parent_model(self, source)
            if not parent_model:
                # if no parent model skip
                continue
            # replace the dot syntac with underscore for getting child count
            status_value = getattr(parent_model, field)
            if utils.is_flag_field_set_for_status(status_value, flag):
                # if it is set then check the count of the child model
                # if only 1 or less left unset the flag
                qs = utils.get_django_query_string_from_source(source)
                child_count = self.__class__.objects.filter(**{qs: parent_model.id}).count()
                if child_count <= 1:
                    status_value = utils.unset_flag_field_for_status(status_value, flag)
                    setattr(parent_model, field, status_value)
                    parent_model.save(update_fields=[field])
        return super(ChildBitfieldModelMixin, self).delete(*args, **kwargs)


class ParentBitfieldModel(ParentBitfieldModelMixin, models.Model):
    pass



try:
    from bitfield.types import Bit, BitHandler
except:
    Bit = None
    BitHandler = None


# AND
def is_flag_set_for_status(status, flag):
    # this is because checking for equality on bithandlers will return false
    # when compared to an int of the same value
    if BitHandler and type(status) == BitHandler:
        status = int(status)
    return (status & flag) == flag


def is_flag_field_set_for_status(status, flag_field):
    if Bit and type(flag_field) == Bit:
        return is_flag_set_for_status(status, flag_field.mask)
    return is_flag_set_for_status(status, (1 << flag_field))


def unset_flag_field_for_status(status, flag_field):
    if Bit and type(flag_field) == Bit:
        return unset_flag_for_status(status, flag_field.mask)
    return unset_flag_for_status(status, (1 << flag_field))


def unset_flag_for_status(status, flag):
    return status & ~flag


# OR
def set_flag_for_status(status, flag):
    return status | flag


def set_flag_field_for_status(status, flag_field):
    if Bit and type(flag_field) == Bit:
        # django-bitfield has the field already shifted
        return set_flag_for_status(status, flag_field.mask)
    return set_flag_for_status(status, (1 << flag_field))


def get_all_related_bitfield_models(model, all_models=[], search_depth=1, current_level=0):
    # recursive function to get models multiple levels deep
    new_models = [f.related_model for f in model._meta.get_fields() if
                  f.auto_created and not f.concrete and hasattr(f.related_model, 'BitfieldMeta')]
    if search_depth == 1:
        return new_models

    for m in new_models:
        all_models.append(m)
        if current_level <= search_depth:
            current_level += 1
            get_all_related_bitfield_models(m, all_models, search_depth=search_depth, current_level=current_level)
    return all_models


def get_field_from_source(source):
    return source.split('.')[-1]


def get_django_query_string_from_source(source):
    return "__".join(source.split(".")[:-1])


def get_parent_model(instance, key_string):
    keys = key_string.split('.')
    for key in keys[:-1]:
        if key in [f.name for f in instance._meta.many_to_many]:
            return None
        instance = getattr(instance, key)
    return instance


def check_and_set_flag(parent_model, field, flag):
    status_value = getattr(parent_model, field)
    if not is_flag_field_set_for_status(status_value, flag):
        status_value = set_flag_field_for_status(status_value, flag)
        setattr(parent_model, field, status_value)
        parent_model.save(update_fields=[field])
